/*
 * copyright (c)2009 http://wamsoft.jp
 * zlib license
 */
#include "sqthread.h"
#include <string.h>

extern SQRESULT sqstd_loadmemory(HSQUIRRELVM v, const char *dataBuffer, int dataSize, const SQChar *filename, SQBool printerror);

namespace sqobject {

// p[^s
static SQRESULT ERROR_INVALIDPARAM(HSQUIRRELVM v) {
	return sq_throwerror(v, _SC("invalid param"));
}

// Xbh݂Ȃ
static SQRESULT ERROR_NOTHREAD(HSQUIRRELVM v) {
	return sq_throwerror(v, _SC("no thread"));
}

// fork Ɏs
static SQRESULT ERROR_FORK(HSQUIRRELVM v) {
	return sq_throwerror(v,_SC("failed to fork"));
}

bool
Thread::isWait()
{
	return !_waitSystem.isNull() || _waitList.len() > 0 || _waitTimeout >= 0;
}

/**
 * @return YXbhƌ݊Ǘ̃XbhvĂ true
 */
bool
Thread::isSameThread(HSQUIRRELVM v)
{
	return _thread.isSameThread(v);
}

/**
 * IuWFNgɑ΂҂
 * @param target ҂Ώ
 * @return YIuWFNg҂Ăꍇ true
 */
bool
Thread::_notifyObject(ObjectInfo &target)
{
	bool find = false;
	if (!_waitSystem.isNull() && _waitSystem == target) { // systemR}hp̑҂
		find = true;
		Thread *th = _waitSystem.getThread();
		if (th) {
			_waitResult = th->_exitCode;
		}
		_waitSystem.clear();
	} else {
		SQInteger i = 0;
		SQInteger max = _waitList.len();
		while (i < max) {
			ObjectInfo obj = _waitList.get(i);
			if (obj == target) {
				find = true;
				_waitResult = obj;
				_waitList.remove(i);
				max--;
			} else {
				i++;
			}
		}
	}
	if (find) {
		_clearWait();
	}
	return find;
}

/**
 * gKɑ΂҂
 * @param name gK
 * @return YIuWFNg҂Ăꍇ true
	 */
bool
Thread::_notifyTrigger(const SQChar *name)
{
	bool find = false;
	int i = 0;
	SQInteger max = _waitList.len();
	while (i < max) {
		ObjectInfo obj = _waitList.get(i);
		if (obj.isSameString(name)) {
			find = true;
			_waitResult = obj;
			_waitList.remove(i);
			max--;
		} else {
			i++;
		}
	}
	if (find) {
		_clearWait();
	}
	return find;
}

// RXgN^
Thread::Thread() : _currentTick(0), _fileHandler(NULL), _waitTimeout(-1), _status(THREAD_NONE)
{
	_waitList.initArray();
}

// RXgN^
Thread::Thread(HSQUIRRELVM v) : Object(v), _currentTick(0), _fileHandler(NULL), _waitTimeout(-1), _status(THREAD_NONE)
{
	_waitList.initArray();
	// s
	if (sq_gettop(v) >= 3) {
		_exec(v,3);
		_entryThread(v);
	}
}

// fXgN^
Thread::~Thread()
{
	_exit();
	_thread.clear();
	_waitList.clear();
}

/**
 * j
 */
void
Thread::_init()
{
	HSQUIRRELVM gv = getGlobalVM();
	sq_newthread(gv, 1024);
	_thread.getStack(gv, -1);
	sq_pop(gv, 1);
}

/**
 * j
 */
void
Thread::_clear()
{
	_clearWait();
	if (_fileHandler) {
		sqobjCloseFile(_fileHandler);
		_fileHandler = NULL;
		_scriptName.clear();
	}
	_args.clear();
	_status = THREAD_NONE;
}

void
Thread::_exit()
{
	notifyAll();
	_clear();
}

/**
 * IuWFNgɑ΂҂NA
 * @param status LZ̏ꍇ true
 */
void
Thread::_clearWait()
{
	// systemp wait̉
	_waitSystem.removeWait(self);
	_waitSystem.clear();

	// ̑ wait ̉
	SQInteger max = _waitList.len();
	for (SQInteger i=0;i<max;i++) {
		_waitList.get(i).removeWait(self);
	}
	_waitList.clearData();
	// ^CAEgw̉
	_waitTimeout = -1;
}

/**
 * p:fork BXbhP VMPUSH
 * @param v squirrelVM
 * @return  true
 */
bool
Thread::_fork(HSQUIRRELVM v)
{
	// XbhIuWFNg̓O[oɐ
	SQInteger max = sq_gettop(v);
	HSQUIRRELVM gv = getGlobalVM();
	sq_pushroottable(gv); // root
	sq_pushstring(gv, SQTHREADNAME, -1);
	if (SQ_SUCCEEDED(sq_get(gv,-2))) { // class
		sq_pushroottable(gv); // :self(root)
		sq_pushnull(gv);      // :delegate
		// Rs[
		int argc = 2;
		if (gv == v) {
			for (int i=2;i<=max;i++) {
				sq_push(gv,i);
				argc++;
			}
		} else {
			for (int i=2;i<=max;i++) {
				sq_move(gv, v, i);
				argc++;
			}
		}
		if (SQ_SUCCEEDED(sq_call(gv, argc, SQTrue, SQTrue))) { // RXgN^Ăяo
			if (gv == v) {
				sq_remove(gv, -2); // class
				sq_remove(gv, -2); // root
			} else {
				sq_move(v, gv, sq_gettop(gv)); // VM̂قɈڂ
				sq_pop(gv, 3); // thread,class,root
			}
			return true;
		}
		sq_pop(gv, 1); // class
	}
	sq_pop(gv,1); // root
	return false;
}

/**
 * p: wait
 * @param v squirrelVM
 * @param idx Y idx ȍ~ɂ̂҂
 */
void
Thread::_wait(HSQUIRRELVM v, int idx)
{
	_clearWait();
	_waitResult.clear();
	SQInteger max = sq_gettop(v);
	for (int i=idx;i<=max;i++) {
		switch (sq_gettype(v, 2)) {
		case OT_INTEGER:
		case OT_FLOAT:
			// l̏ꍇ̓^CAEg҂
			{
				SQInteger timeout;
				sq_getinteger(v, i, &timeout);
				if (timeout >= 0) {
					if (_waitTimeout < 0  || _waitTimeout > timeout) {
						_waitResult.getStack(v, i);
						_waitTimeout = timeout;
					}
				}
			}
			break;
		case OT_STRING:
			// ҂Xgɓo^
			_waitList.append(ObjectInfo(v,i));
			break;
		case OT_INSTANCE:
			// IuWFNgɑ҂o^Ă҂Xgɓo^
			{
				ObjectInfo o;
				o.getStackWeak(v,i);
				o.addWait(self);
				_waitList.append(o);
			}
			break;
		default:
			break;
		}
	}
}

/**
 * p: system̑҂o^BX^bN擪ɂXbh҂
 * @param v squirrelVM
 */
void
Thread::_system(HSQUIRRELVM v)
{
	_clearWait();
	_waitResult.clear();
	_waitSystem.getStackWeak(v, -1);
	_waitSystem.addWait(self);
}

// ---------------------------------------------------------------

SQRESULT
Thread::wait(HSQUIRRELVM v)
{
	_wait(v);
	return 0;
}

/**
 * wait̃LZ
 */
void
Thread::cancelWait()
{
	_clearWait();
	_waitResult.clear();
}

/**
 * p: exec
 * @param v squirrelVM
 * @param idx ̃CfbNXɂ̂sJnBȂXNvgAt@NVȂ璼
 */
void
Thread::_exec(HSQUIRRELVM v, int idx)
{
	_clear();
	_thread.clear();
	// Xbh擪ɃXNvg[h
	if (sq_gettype(v, idx) == OT_STRING) {
		// XNvgwŒx[h
		_scriptName.getStack(v, idx);
		_fileHandler = sqobjOpenFile(getString(v, idx));
		_status = THREAD_LOADING_FILE;
	} else {
		// t@NVw
		_func.getStack(v, idx);
		_status = THREAD_LOADING_FUNC;
	}

	// L^
	SQInteger max = sq_gettop(v);
	if (max > idx) {
		_args.initArray();
		for (int i=idx+1;i<=max;i++) {
			_args.append(v, i);
		}
	}
}

/**
 * XbhƂēo^
 */
void
Thread::_entryThread(HSQUIRRELVM v)
{
	// XbhƂēo^
	ObjectInfo thinfo(v,1);
	SQInteger max = threadList.len();
	for (int i=0;i<max;i++) {
		if (threadList.get(i) == thinfo) {
			return;
		}
	}
	newThreadList.append(thinfo);
}


/**
 * sJn
 * @param func sΏۃt@NVB̏ꍇYXNvgǂݍ
 */
SQRESULT
Thread::exec(HSQUIRRELVM v)
{
	if (sq_gettop(v) <= 1) {
		return ERROR_INVALIDPARAM(v);
	}
		
	_exec(v);
	_entryThread(v);

	return SQ_OK;
}

/**
 * sI
 */
SQRESULT
Thread::exit(HSQUIRRELVM v)
{
	if (sq_gettop(v) >= 2) {
		_exitCode.getStack(v,2);
	} else {
		_exitCode.clear();
	}
	_exit();
	return SQ_OK;
}

/**
 * @return sXe[^X
 */
SQRESULT
Thread::getExitCode(HSQUIRRELVM v)
{
	_exitCode.push(v);
	return 1;
}

/**
 * s~
 */
void
Thread::stop()
{
	if (_status == THREAD_RUN) {
		_status = THREAD_STOP;
	}
}

/**
 * sĊJ
 */
void
Thread::run()
{
	if (_status == THREAD_STOP) {
		_status = THREAD_RUN;
	}
}

/**
 * @return sXe[^X
 */
int
Thread::getStatus()
{
	return isWait() ? THREAD_WAIT : _status;
}

/**
 * Xbh̃C
 * @param diff oߎ
 * @return XbhsIȂ true
 */
bool
Thread::_main(long diff)
{
	// XbhƂēłĂȂꍇ͑I
	if (_status == THREAD_NONE) {
		return true;
	}

	if (_status == THREAD_LOADING_FILE) {
		// t@Cǂݍݏ
		const char *dataAddr;
		int dataSize;
		if (sqobjCheckFile(_fileHandler, &dataAddr, &dataSize)) {
			_init();
			SQRESULT ret = sqstd_loadmemory(_thread, dataAddr, dataSize, _scriptName.getString(), SQTrue);
			sqobjCloseFile(_fileHandler);
			_fileHandler = NULL;
			if (SQ_SUCCEEDED(ret)) {
				_status = THREAD_RUN;
			} else {
				// exit
				printError();
				_exit();
				return true;
			}
		} else {
			// ǂݍ݊҂
			return false;
		}
	} else if (_status == THREAD_LOADING_FUNC) {
		// XNvgǂݍݏ
		_init();
		_func.push(_thread);
		_func.clear();
		_status = THREAD_RUN;
	}

	_currentTick += diff;
	
	// ^CAEg
	if (_waitTimeout >= 0) {
		_waitTimeout -= diff;
		if (_waitTimeout < 0) {
			_clearWait();
		}
	}
		
	// Xbhs
	if (!isWait() && _status == THREAD_RUN) {
		SQRESULT result;
		if (sq_getvmstate(_thread) == SQ_VMSTATE_SUSPENDED) {
			_waitResult.push(_thread);
			_waitResult.clear();
			result = sq_wakeupvm(_thread, SQTrue, SQFalse, SQTrue, SQFalse);
		} else {
			sq_pushroottable(_thread);
			SQInteger n = _args.pushArray(_thread) + 1;
			_args.clear();
			result = sq_call(_thread, n, SQFalse, SQTrue);
		}
		if (SQ_FAILED(result)) {
			// XbhG[I
			printError();
			_exit();
		} else  if (sq_getvmstate(_thread) == SQ_VMSTATE_IDLE) {
			// XbhɏI
			if (result > 0) {
				_exitCode.getStack(_thread, -1);
			}
			_exit();
		}
	}
	
	return _status == THREAD_NONE;
}

// -------------------------------------------------------------------------

/**
 * Xbh̃G[̕\
 */
void
Thread::printError()
{
	SQPRINTFUNCTION print = sq_getprintfunc(_thread);
	if (print) {
		sq_getlasterror(_thread);
		const SQChar *err;
		if (SQ_FAILED(sq_getstring(_thread, -1, &err))) {
			err = _SC("unknown");
		}
		print(_thread,_SC("error:%s:%s\n"), _scriptName.getString(), err);
		sq_pop(_thread, 1);
	}
}

/**
 * Xbh̔j
 */
void
Thread::init()
{
	threadList.initArray();
	newThreadList.initArray();
}

/*
 * ԍXV
 * @param diff oߎ
 */
void
Thread::update(long diff)
{
	diffTick = diff;
	currentTick += diff;
}

/*
 * sC[v
 * ݑ݂Xbh𑍂Ȃ߂łPxsB
 * VXe{̂̃C[v(Cxg{摜)
 * 1xĂяoƂŋ@\Bꂼ̃Xbh́A
 * 疾I suspend() ܂ waitñ\bhĂяoď
 * ̃XbhɈϏKvB
 * @return 쒆̃Xbh̐
 */
int
Thread::main()
{
	SQInteger i=0;
	SQInteger max = threadList.len();
	while (i < max) {
		Thread *th = threadList.get(i).getThread();
		if (!th || th->_main(diffTick)) {
			threadList.remove(i);
			max--;
		} else {
			i++;
		}
	}
	threadList.appendArray(newThreadList);
	newThreadList.clearData();
	return (int)threadList.len();
};

int
Thread::getThreadCount()
{
	return (int)threadList.len();
}

/**
 * XNvgsJnp
 * @param scriptName XNvg
 * @param argc ̐
 * @param argv 
 * @return Ȃ true
 */
bool
Thread::fork(const SQChar *scriptName, int argc, const SQChar **argv)
{
	HSQUIRRELVM gv = getGlobalVM();
	sq_pushroottable(gv); // root
	sq_pushstring(gv, SQTHREADNAME, -1);
	if (SQ_SUCCEEDED(sq_get(gv,-2))) { // class
		sq_pushroottable(gv); // :self(root)
		sq_pushnull(gv);      // :delegate
		sq_pushstring(gv, scriptName, -1); // :func
		int n = 3;
		for (int i=0;i<argc;i++) {
			sq_pushstring(gv, argv[i], -1);
			n++;
		}
		if (SQ_SUCCEEDED(sq_call(gv, n, SQTrue, SQTrue))) { // RXgN^Ăяo
			sq_pop(gv, 3); // thread,class,root
			return true;
		}
		sq_pop(gv, 1); // class
	}
	sq_pop(gv,1); // root
	return false;
}

/**
 * Xbh̔j
 */
void
Thread::done()
{
	// SXbhf
	threadList.appendArray(newThreadList);
	newThreadList.clearData();
	SQInteger max = threadList.len();
	for (SQInteger i=0;i<max;i++) {
		Thread *th = threadList.get(i).getThread();
		if (th) { th->_exit();}
	}
	threadList.clearData();
	// Xĝj
	threadList.clear();
	newThreadList.clear();
}

// -------------------------------------------------------------
// O[oXbhp
// -------------------------------------------------------------

ObjectInfo Thread::threadList; //< Xbhꗗ
ObjectInfo Thread::newThreadList; //< VKXbhꗗ
long Thread::currentTick = 0;  //< ݂̃VXetick
long Thread::diffTick = 0;  //< ̌Ăяo

// -------------------------------------------------------------
// O[o\bhp
// -------------------------------------------------------------

/**
 * ݎ̎擾
 */
SQRESULT
Thread::global_getCurrentTick(HSQUIRRELVM v)
{
	sq_pushinteger(v, currentTick);
	return 1;
}

/**
 * ̎擾
 */
SQRESULT
Thread::global_getDiffTick(HSQUIRRELVM v)
{
	sq_pushinteger(v, diffTick);
	return 1;
}

/*
 * @return ݂̃XbhԂ
 */
SQRESULT
Thread::global_getCurrentThread(HSQUIRRELVM v)
{
	SQInteger max = threadList.len();
	for (SQInteger i=0;i<max;i++) {
		ObjectInfo obj = threadList.get(i);
		Thread *th = obj.getThread();
		if (th && th->isSameThread(v)) {
			obj.push(v);
			return 1;
		}
	}
	return ERROR_NOTHREAD(v);
}

/*
 * @return ݂̃XbhꗗԂ
 */
SQRESULT
Thread::global_getThreadList(HSQUIRRELVM v)
{
	threadList.pushClone(v);
	return 1;
}

/*
 * XNvgVXbhƂĎs
 *  return Thread(func); 
 * @param func XbhŎst@NV
 * @return VXbh
 */
SQRESULT
Thread::global_fork(HSQUIRRELVM v)
{
	if (!_fork(v)) {
		return ERROR_FORK(v);
	}
	return 1;
}

/**
 * @return ݎs̃XbhIuWFNg(Thread*)
 */
Thread *
Thread::getCurrentThread(HSQUIRRELVM v)
{
	SQInteger max = threadList.len();
	for (SQInteger i=0;i<max;i++) {
		ObjectInfo obj = threadList.get(i);
		Thread *th = obj.getThread();
		if (th && th->isSameThread(v)) {
			return th;
		}
	}
	return NULL;
}

/**
 * XNvg؂ւ
 * @param func XbhŎst@NV
 */
SQRESULT
Thread::global_exec(HSQUIRRELVM v)
{
	Thread *th = getCurrentThread(v);
	if (!th) {
		return ERROR_NOTHREAD(v);
	}
	if (sq_gettop(v) <= 1) {
		return ERROR_INVALIDPARAM(v);
	}
	th->_exec(v);
	return sq_suspendvm(v);
}

/**
 * sXbh̏I
 */
SQRESULT
Thread::global_exit(HSQUIRRELVM v)
{
	Thread *th = getCurrentThread(v);
	if (!th) {
		return ERROR_NOTHREAD(v);
	}
	th->exit(v);
	return sq_suspendvm(v);
}

/**
 * R}hs
 * @param func XbhŎst@NV
 * @return IR[h
 */
SQRESULT
Thread::global_system(HSQUIRRELVM v)
{
	Thread *th = getCurrentThread(v);
	if (!th) {
		return ERROR_NOTHREAD(v);
	}
	if (!_fork(v)) {
		return ERROR_FORK(v);
	}
	th->_system(v);
	sq_pop(v,1);
	return sq_suspendvm(v);
}

/**
 * sXbh̏҂
 * @param target int:ԑ҂(ms), string:gK҂, obj:IuWFNg҂
 * @param timeout ^CAEg(ȗ͖ɑ҂)
 * @return ҂LZꂽ true
 */
SQRESULT
Thread::global_wait(HSQUIRRELVM v)
{
	Thread *th = getCurrentThread(v);
	if (!th) {
		return ERROR_NOTHREAD(v);
	}
	th->_wait(v);
	return sq_suspendvm(v);
}

/**
 * SXbhւ̃gKʒm
 * @param name ҂gK
 */
SQRESULT
Thread::global_trigger(HSQUIRRELVM v)
{
	const SQChar *name = getString(v, 2);
	SQInteger max = threadList.len();
	for (SQInteger i=0;i<max;i++) {
		ObjectInfo obj = threadList.get(i);
		obj.notifyTrigger(name);
	}
	return SQ_OK;
}

/**
 * O[o\bho^
 */
void
Thread::registerGlobal()
{
	// \bho^iOj
#define REGISTERMETHODNAME(name, method) \
	sq_pushstring(v, _SC(#name), -1);\
	sq_newclosure(v, method, 0);\
	sq_createslot(v, -3);

	// enum o^iOj
#define REGISTERENUM(name, value) \
	sq_pushstring(v, _SC(#name), -1); /* O push */ \
	sq_pushinteger(v, value);          /* l push */ \
	sq_createslot(v, -3)              /* e[uɓo^ */
	
	HSQUIRRELVM v = getGlobalVM();

	// O[o\bh̓o^
	sq_pushroottable(v); // root
	REGISTERMETHODNAME(getCurrentTick, global_getCurrentTick);
	REGISTERMETHODNAME(getDiffTick, global_getDiffTick);
	REGISTERMETHODNAME(getCurrentThread, global_getCurrentThread);
	REGISTERMETHODNAME(getThreadList, global_getThreadList);
	REGISTERMETHODNAME(fork, global_fork);
	REGISTERMETHODNAME(exec, global_exec);
	REGISTERMETHODNAME(exit, global_exit);
	REGISTERMETHODNAME(system, global_system);
	REGISTERMETHODNAME(wait, global_wait);
	REGISTERMETHODNAME(notify, global_trigger);
	sq_pop(v, 1); // root
	
	// 萔̓o^
	sq_pushconsttable(v); // consttable
	sq_pushstring(v, _SC("THREADSTATUS"), -1); // e[u push
	sq_newtable(v);                  // V enum e[u
	REGISTERENUM(NONE,THREAD_NONE);
	REGISTERENUM(LOADING_FILE,THREAD_LOADING_FILE);
	REGISTERENUM(LOADING_FUNC,THREAD_LOADING_FUNC);
	REGISTERENUM(STOP,THREAD_STOP);
	REGISTERENUM(RUN,THREAD_RUN);
	REGISTERENUM(WAIT,THREAD_WAIT);
	sq_createslot(v, -3);              /* e[uɓo^ */
	sq_pop(v, 1); // consttable
}

};
